import re

RSS_FEED_FILE_NAME = '.index.php'
RSS_FEED_INFO_EXTENSION = 'info'
RSS_FEED_SHOW_INDEX = 'index'
RSS_FEED_SHOW_IMAGE = 'image.jpg'
RSS_FEED_VERSION = '$SPODCAST_VERSION$ '
VERSION_NOT_FOUND = 0

def get_index_version(filename) -> str:
    with open(filename, 'rb') as f:
        for line in f.readlines():
            m = re.search(RSS_FEED_VERSION + ' (\d+.\d+.\d+)', str(line))
            if m:
                return int(m[1].replace('.',''))

    return VERSION_NOT_FOUND


def RSS_FEED_CODE(version):
    return r'''<?php
/* ''' + RSS_FEED_VERSION + version + r''' */
const SHOW_INDEX = "''' + RSS_FEED_SHOW_INDEX + r'''";
const INFO = "''' + RSS_FEED_INFO_EXTENSION + r'''";
$PROTOCOL = (empty($_SERVER['HTTPS'])) ? "http://" : "https://";
header("Content-type: text/xml");
$feed_name = "Spodcast autofeed";
$feed_description = "Spodcast autofeed";
$base_url = strtok($PROTOCOL . $_SERVER['HTTP_HOST'] . $_SERVER['REQUEST_URI'], '?');
$feed_logo = "$base_url/''' + RSS_FEED_SHOW_IMAGE + r'''";
$feed_link = $base_url;
$allowed_extensions = array('mp4','m4a','aac','mp3','ogg');

$sinfo=SHOW_INDEX.".".INFO;
if(file_exists($sinfo)) {
    $json=file_get_contents($sinfo);
    $info=json_decode($json);
    $feed_name=$info->title;
    $feed_description=$info->description;
    $feed_link=$info->link;
}

?>
<?php echo '<?xml version="1.0" encoding="utf-8"?>';  // use php to output the "<?" ?>
<rss version="2.0" xmlns:atom="http://www.w3.org/2005/Atom" xmlns:media="http://search.yahoo.com/mrss/" xmlns:itunes="http://www.itunes.com/dtds/podcast-1.0.dtd" >
    <channel>
        <title><?php echo $feed_name; ?></title>
        <link><?php echo $feed_link; ?></link>
        <image>
            <url><?php echo $feed_logo; ?></url>
            <title><?php echo $feed_name; ?></title>
            <link><?php echo $feed_link; ?></link>
        </image>
        <description><?php echo $feed_description; ?></description>
        <atom:link href="<?php echo $base_url; ?>" rel="self" type="application/rss+xml" />
<?php
$raw_files = scandir ('.');
usort($raw_files, function($a, $b) {
    return filemtime($a) < filemtime($b);
});

foreach ($raw_files as &$raw_file) {
    $raw_file_info = pathinfo($raw_file);
    $extension = strtolower($raw_file_info['extension']);
    if(!empty($extension)) {
        if(in_array ($extension,$allowed_extensions)) {
            $finfo=$raw_file.".".INFO;
            if(file_exists($finfo)) {
                $json=file_get_contents($finfo);
                $info=json_decode($json);
                echo "        <item>\n";
                echo "            <title>".$info->title."</title>\n";
                echo "            <description>".$info->description."</description>\n";
                echo "            <guid>".$info->guid."</guid>\n";
                echo "            <link>".$base_url.$info->filename."</link>\n";
                echo "            <enclosure url=\"".$base_url.$info->filename."\" length=\"".$info->size."\" type=\"".$info->mimetype."\" />\n";
                echo "            <media:content url=\"".$base_url.$info->filename."\" medium=\"".$info->medium."\" duration=\"".$info->duration."\" type=\"".$info->mimetype."\" />\n";
                echo "            <pubDate>".$info->date."</pubDate>\n";
                echo "            <itunes:duration>".$info->duration."</itunes:duration>\n";
                echo "        </item>\n";
            }
        }
    }
}
?>
    </channel>
</rss>'''

def RSS_INDEX_CODE(bin_path, config_name, version):
    return r'''<?php
/* ''' + RSS_FEED_VERSION + version + r''' */
const INFO="''' + RSS_FEED_INFO_EXTENSION + r'''";
const SHOW_INFO="''' + RSS_FEED_SHOW_INDEX + r'''.".INFO;
const SPODCAST="''' + bin_path + r'''";
const SPODCAST_CONFIG="''' + config_name + r'''";
const SHOW_IMAGE="''' + RSS_FEED_SHOW_IMAGE + r'''";
const FEEDS_INFO="feeds.".INFO;
const SETTINGS_INFO="settings.".INFO;
const MAX_EPISODES=3;
const KEEP_EPISODES=5;
const UPDATEABLE=['max','keep'];
const CLI_COMMANDS=['refresh'];
const SUCCESS='success';
const ERROR='error';
const LOG_LEVEL='warning';
const NOT_FOUND=-1;

$SPODCAST_CONFIG=dirname(__FILE__)."/".SPODCAST_CONFIG;
$SPODCAST_COMMAND=SPODCAST." -c ".$SPODCAST_CONFIG;

# CLI
if (PHP_SAPI == "cli") {
    global $SPODCAST_CONFIG;
    if (count($argv) < 2) {
        echo "use: php ".__FILE__." <command> [options...]".PHP_EOL;
        echo "     commands: ".implode("|",CLI_COMMANDS).PHP_EOL;
        die();
    }

    $settings=read(dirname(__FILE__)."/".SETTINGS_INFO);
    $feeds=read(dirname(__FILE__)."/".FEEDS_INFO);

    $command=$argv[1];
    if($command == "refresh") {
        list($retval, $result) = refresh_shows($feeds);
        if ($retval > 0) {
            echo "An error occurred during refresh, return value was $retval" . PHP_EOL;
        }
        echo implode(PHP_EOL, $result);
    }
    die();
}

# CGI/API
$PROTOCOL = (empty($_SERVER['HTTPS'])) ? "http://" : "https://";
$SPODCAST_URL = $PROTOCOL . $_SERVER['HTTP_HOST'] . parse_url($_SERVER['REQUEST_URI'], PHP_URL_PATH);
$feeds=get_feeds(dirname(__FILE__));
$settings=get_settings();
$config=get_spodcast_config();
$ERROR_MESSAGE=null;
$ERROR_DETAILS=null;

function get_feeds($dir) {
    global $SPODCAST_URL;
    foreach(glob($dir."/*/".SHOW_INFO) as $show_info) {
        $episodes=get_episodes(dirname($show_info));
        $json=file_get_contents($show_info);
        $info=json_decode($json);
        $feeds[$info->link]['title']=$info->title;
        $feeds[$info->link]['image']=$info->image;
        $feeds[$info->link]['episodes']=count($episodes);
        $feeds[$info->link]['last']=date("Y-m-d",strtotime($episodes[0]['date']));
        $feeds[$info->link]['directory']=dirname($show_info);
        $feeds[$info->link]['max']=$info->max ?? MAX_EPISODES;
        $feeds[$info->link]['keep']=$info->keep ?? KEEP_EPISODES;
        $feeds[$info->link]['feed']=$SPODCAST_URL.basename(dirname($show_info));
    }
    uasort($feeds, fn ($a, $b) => strnatcmp($a['title'], $b['title']));
    store($feeds, FEEDS_INFO);
    return $feeds;
}

function get_episodes($dir) {
    $episodes=[];
    foreach(glob($dir."/*.".INFO) as $episode_info) {
        if(basename($episode_info) == SHOW_INFO) {
            continue;
        }
        $json=file_get_contents($episode_info);
        $info=json_decode($json);
        $episodes[]=["filename"=>$info->filename,"date"=>$info->date,"title"=>$info->title];
    }
    usort($episodes, fn ($a, $b) => strtotime($b["date"]) - strtotime($a["date"]));
    return $episodes;
}

function get_settings() {
    global $SPODCAST_URL;
    $settings=read(SETTINGS_INFO);
    $settings['spodcast_url']=$settings['spodcast_url'] ?? $SPODCAST_URL;
    $settings['update_start']=$settings['update_start'] ?? 0;
    $settings['update_rate']=$settings['update_rate'] ?? 1;
    $settings['update_enabled']=$settings['update_enabled'] ?? false;
    $settings['log_level']=$settings['log_level'] ?? LOG_LEVEL;
    store($settings, SETTINGS_INFO);
    return $settings;
}

function get_spodcast_config() {
    global $SPODCAST_CONFIG;
    $config=read($SPODCAST_CONFIG);
    return $config;
}

function get(&$var, $default=null) {
    return isset($var) ? $var : $default;
}

function read($file) {
    if(is_readable($file)) {
        $json=file_get_contents($file);
        $info=json_decode($json, true);
    } else {
        $info=[];
    }
    return $info;
}

function store($info, $file) {
    $f = fopen($file,'w');
    $result = fwrite($f, json_encode($info));
    fclose($f);
    return ($result === false) ? ERROR : SUCCESS;
}

function cron_signature($crontab, $CRON_SIGNATURE) {
    $index = 0;
    foreach ($crontab as $line) {
        if (strpos($line, $CRON_SIGNATURE)  !== false) {
            return $index;
        }
        $index++;
    }
    return NOT_FOUND;
}

function debug($var) {
    ob_start();
    var_dump($var);
    error_log(ob_get_clean());
}

function submit_crontab($crontab) {
    $retval = null;
    $output = null;
    $tempfile=tempnam(sys_get_temp_dir(), 'spodcast');
    file_put_contents($tempfile, implode(PHP_EOL,$crontab));
    $command="crontab ".$tempfile;
    exec($command, $output, $retval);
    unlink($tempfile);

    return [$retval, $output];
}

function get_range($start, $rate) {
    for ($i=0; $i < $rate; $i++) {
        $arr[]=(($start%(24/$rate))+($i*(24/$rate)))%24;
    }
    return implode(",", $arr);
}

function background_check_id($id) {
    $runfile = md5($id).".json";
    $log = md5($id).".log";
    if (is_readable($runfile)) {
        $info = read($runfile);
        if (background_check($info['pid'])) {
            return true;
        }
        unlink($runfile);
        unlink($log);
    }
    return false;
}

function background_check($pid) {
    try {
        $result = shell_exec(sprintf("ps %d", $pid));
        if (count(preg_split("/\n/", $result)) > 2) {
            return true;
        }
    } catch(Exception $e) {}
    return false;
}

# [ status, output ] return functions
function update_scheduler($enable, $start, $rate) {
    $CRON_SIGNATURE="SPODCAST:".dirname(__FILE__);
    $crontab=null;
    $retval=null;
    exec("crontab -l", $crontab, $retval);
    if ($retval == 0) {
        $index=cron_signature($crontab, $CRON_SIGNATURE);
        if ($enable == true) {
            if($index !== NOT_FOUND) {
                array_splice($crontab, $index, 1);
            }
            $crontab[]=sprintf("%d %s * * * php %s refresh # %s".PHP_EOL, rand(5,25), get_range($start, $rate),  __FILE__, $CRON_SIGNATURE);
            return submit_crontab($crontab);
        } else {
            if ($index !== NOT_FOUND) {
                array_splice($crontab, $index, 1);
                $crontab[count($crontab)-1]=rtrim($crontab[count($crontab)-1]).PHP_EOL;
                return submit_crontab($crontab);
            }
        }
    } else {
        return [$retval, "failed to update scheduler"];
    }
}

function login($username, $password, $return_output=false) {
    global $SPODCAST_COMMAND;
    global $settings;
    $output = null;
    $retval = null;
    $tempfile=tempnam(sys_get_temp_dir(), 'spodcast');
    file_put_contents($tempfile, "$username $password");
    $command=$SPODCAST_COMMAND . " --log-level " . $settings['log_level'] . " -l ".$tempfile." 2>&1";
    exec($command, $output, $retval);
    unlink($tempfile);

    return [$retval, $output];
}

function background_run($command, $id) {
    $retval=null;
    $md5 = md5($id);
    $runfile = $md5.".json";
    $output = $md5.".log";
    $cmd = sprintf("nohup %s > %s 2>&1 & echo $!", $command, $output);
    exec($cmd, $pid, $retval);
    if ($retval == 0 && count($pid) > 0 && $pid[0] > 0) {
        $info['command']=$command;
        $info['log']=$output;
        $info['pid']=(int) $pid[0];
        store($info, $runfile);
        return [$retval, $pid[0]];
    } else {
        return [$retval, 0];
    }
}

function background_add_feed($url, $max) {
    global $SPODCAST_COMMAND;
    global $settings;
    $output = null;
    $retval = null;
    $command=$SPODCAST_COMMAND  . " --log-level " . $settings['log_level'] . " --max-episodes ".(int)$max." ".escapeshellarg($url);
    list($retval, $pid) = background_run($command, $url);
    if ($retval == 0 && (int) $pid > 0) {
        return [$retval, $pid];
    }
    return [$retval, 0];
}


function add_feed($url, $max) {
    global $SPODCAST_COMMAND;
    global $settings;
    $output = null;
    $retval = null;
    $command=$SPODCAST_COMMAND  . " --log-level " . $settings['log_level'] . " --max-episodes ".(int)$max." ".escapeshellarg($url)." 2>&1";
    exec($command, $output, $retval);
    return [$retval, $output];
}

function update_feed($url, $max, $keep, $feeds) {
    $output = null;
    $retval = null;
    if ($max > 0) {
    list($retval,$output) = add_feed($url, $max);
        if ($retval > 0) {
            return [$retval, $output];
        }
    }
    $directory=$feeds[$url]['directory'];
    $episodes=get_episodes($directory);
    if(count($episodes) > $keep) {
        $to_delete=array_splice($episodes, $keep);
        foreach ($to_delete as $episode) {
            $command="rm -f ".escapeshellarg($directory."/".$episode['filename'])." 2>&1";
            exec($command, $output, $retval);
            if ($retval > 0) {
                return [$retval, $output];
            }
            $command="rm -f ".escapeshellarg($directory."/".$episode['filename']).".".INFO." 2>&1";
            exec($command, $output, $retval);
            if ($retval > 0) {
                return [$retval, $output];
            }
        }
    }
    return [$retval, $output];
}

function delete_feed($url, $return_output=false) {
    $output = null;
    $retval = null;
    $feeds=read(FEEDS_INFO);
    $feed_dir=$feeds[$url]['directory'];
    $command="rm -rf ".$feed_dir." 2>&1";
    exec($command, $output, $retval);
    return [$retval, $output];
}

function update_show($url, $field, $value) {
    if (array_search($field, UPDATEABLE) === false) {
        return [1, "$field is not an updateable field"];
    } else {
        $feeds=read(FEEDS_INFO);
        $show_dir=$feeds[$url]['directory'];
        $show=read($show_dir."/".SHOW_INFO);
        $show[$field]=$value;
        store($show, $show_dir."/".SHOW_INFO);
        return [0, "$field set to $value"];
    }
}

function refresh_shows($feeds) {
    $result = [];
    $status = 0;
    foreach ($feeds as $url => ["title"=>$title, "directory"=>$directory, "max"=>$max, "keep"=>$keep]) {
        $output = null;
        $retval = 0;
        list($retval, $output) = update_feed($url, $max, $keep, $feeds, true);
        if ($retval > 0) {
            $status = $retval;
        }
        $result = array_merge($result, $output);
    }
    return [$status, $result];
}

# terminating functions

function show_error($message, $details) {
    header("Location: ./?action=error&message=".urlencode($message)."&details=".urlencode(implode(PHP_EOL,$details)));
    die();
}

function json_response($info) {
    header('Content-Type: application/json; charset=utf-8');
    echo json_encode($info, true);
    exit();
}

# CGI commands

switch(get($_GET['action'])) {

case 'refresh':
    $url = get($_POST['url']);
    $max = (int) get($_POST['max']) ?? MAX_EPISODES;
    $keep = (int) get($_POST['keep']) ?? KEEP_EPISODES;
    $info['url']=$url;
    $info['show']=$feeds[$url];
    $info['id']=basename($url);
    if ($max > $keep) {
        $info['result'] = "It does not make sense for the number of episodes to refresh to be larger than the number of episodes to keep.";
        $info['status'] = 'ERROR';
    } else {
        list($result,$output) = update_feed($url, $max, $keep, $feeds, false);
        $info['status'] = ($result == 0) ? 'SUCCESS': 'ERROR';
        if ($result != 0) {
            $info['result'] = "Refresh failed";
        }
    }
    json_response($info);

case 'new':
    $url = get($_POST['url']);
    list($result, $output) = background_add_feed($url, MAX_EPISODES, false);
    $info['url']=$url;
    $info['show']=$feeds[$url] ?? null;
    $info['id']=basename($url);
    $info['status']= ($result == 0) ? 'SUCCESS': 'ERROR';
    json_response($info);

case 'update':
    $url = get($_POST['url']);
    $field = get($_POST['field']);
    $value = get($_POST['value']);
    if (in_array(strtolower($field), UPDATEABLE)) {
        list($result, $output) = update_show($url, $field, $value);
        if ($result == 0) {
            $info['result'] = $output;
            $info['status'] = 'SUCCESS';
        } else {
            $info['result'] = "Update failed";
            $info['status'] = 'ERROR';
        }
    } else {
        $info['result'] = "$field can not be updated";
        $info['status'] = 'ERROR';
    }

    json_response($info);

case 'schedule':
    $enable = (get($_POST['enable']) == "true" ? true : false);
    $start = (int) get($_POST['start']);
    $rate = (int) get($_POST['rate']);
    $settings['update_enabled'] = $enable;
    $settings['update_start'] = $start;
    $settings['update_rate'] = $rate;
    $result = store($settings, SETTINGS_INFO);
    if ($result === ERROR) {
        $info['status'] = 'ERROR';
        $info['result'] = 'Could not store scheduler preferences, giving up';
    } else {
        list($result, $output) = update_scheduler($enable, $start, $rate);
        if ($result == 0) {
            $info['status'] = 'SUCCESS';
            $info['result'] = ($enable === true) ? "Scheduled updates enabled, $rate times per day starting at $start:00" : 'Scheduled updates disabled';
        } else {
            $info['status'] = 'ERROR';
            $info['result'] = implode(PHP_EOL, $result);
        }
    }
    json_response($info);

case 'transcode':
    $enable = (get($_POST['enable']) == "true" ? true : false);
    $config['TRANSCODE']=$enable;
    $result = store($config, $SPODCAST_CONFIG);
    if ($result === SUCCESS) {
        $info['status'] = 'SUCCESS';
        $info['result'] = ($enable) ? 'Transcoding enabled' : 'Transcoding disabled';
    } else {
        $info['status'] = 'ERROR';
        $info['result'] = 'Could not enable transcoding: can not write to config file';
    }
    json_response($info);

case 'logging':
    $level = get($_POST['level']) ?? LOG_LEVEL;
    $config['LOG_LEVEL']=$level;
    if (in_array(strtolower($level), ['critical','error','warning','info','debug'])) {
        $result = store($config, $SPODCAST_CONFIG);
        if ($result === SUCCESS) {
            $info['status'] = 'SUCCESS';
            $info['result'] = 'Log level set to '. $level;
        } else {
            $info['status'] = 'ERROR';
            $info['result'] = 'Could not change log level: can not write to Spodcast config file';
        }
    } else {
        $info['status'] = 'ERROR';
        $info['result'] = 'Invalid log level ' . $level;
    }
    json_response($info);

case 'status':
    $url = get($_POST['url']);
    $info['url'] = $url;
    $info['show']=$feeds[$url] ?? null;
    $info['id'] = basename($url);
    $info['status'] = background_check_id($url) ? 'ACTIVE' : 'READY';
    json_response($info);

case 'delete':
    $url = get($_POST['url']);
    $info['url'] = $url;
    $info['id'] = basename($url);
    list($result, $output) = delete_feed($url);
    if ($result !== 0) {
        $info['status'] = 'ERROR';
        $info['result'] = "Delete feed failed: " . implode(PHP_EOL, $output);
    } else {
        $info['status'] = 'SUCCESS';
        $info['result'] = "Deleted <i>" . $feeds[$url]['title'] . "</i>" ;
    }
    json_response($info);

case 'login':
    $username = get($_POST['username']);
    $password = get($_POST['password']);
    list($result, $output) = login($username, $password);
    if ($result === 0) {
        $info['status'] = 'SUCCESS';
        $info['result'] = 'Login succeeded';
    } else {
        $info['status'] = 'ERROR';
        $info['result'] = 'Login failed for user '. $username;
    }
    json_response($info);

case 'update_shows':
    list($result, $output) = refresh_shows($feeds);
    $info['status'] = ($result == 0) ? 'SUCCESS' : 'ERROR';
    $info['result'] = $output;
    json_response($info);

case 'error':
    $ERROR_MESSAGE = get($_POST['message']);
    $ERROR_DETAILS = get($_POST['details']);
default:
    break;
}

$ACTIVE=[];

foreach (array_keys($feeds) as $url) {
    if (background_check_id($url)) {
        $ACTIVE[$url]=true;
    }
}

$TRANSCODE_ENABLED=(array_key_exists('TRANSCODE', $config)) ? (($config['TRANSCODE'] == "True") ? true : false) : false;
$LOG_LEVEL=$config['LOG_LEVEL'];
$UPDATE_ENABLED=$settings['update_enabled'];
$UPDATE_START=$settings['update_start'];
$UPDATE_RATE=$settings['update_rate'];
?>
<html>
<head>
    <title>Spodcast feed manager</title>
    <style>
        div, body, html { background-color: white; margin: 0.5em; }
        div#new-feed { margin: 0em; display: flex; }
        input#url { flex: 1; }
        div#settings-header { text-align: right; }
        div input, div button { height: 2em; font-size: 1.2em; margin: 0 1em 1em 0; }
        div#loading { position: fixed; display: flex; width: 100%; height: 100%; top: 0; left: 0; opacity: 0.8; background-color: #fff; z-index: 99; justify-content: center; align-items: center; font-size: 200%; }
        div#settings, div#error { position: fixed; top: 0; left: 0; width: 100%; height: 100%; z-index: 99; opacity: 0.95; margin: 0.5em auto; }
        div.hidden { visibility: hidden; }
        div.overlay { position: absolute; width: 100%; height: 100%; left: 0; right: 0; background: white; opacity: 0.5; z-index: 100; display: flex; justify-content: center; align-items: center; }
        .show, .setgroup, .error { position: relative; display: flex; flex-wrap: wrap; margin: 0 0 2em 0; padding: 0; font-family: sans-serif; border: 1px solid #F0F0F0; background: #F0F0F0; border-radius: 6px; }
        .show .placeholder { height: 4em; }
        .setgroup { border: 1px solid black; margin: 1em; }
        .error { border: 2px solid red; margin: 1em; }
        .entry, .setitem, .errorline { display: flex; box-sizing: border-box; flex-grow: 1; width: 100%; padding: 0.4em 0.6em; overflow: hidden; align-items: center; text-align: left; background: #F0F0F0; }
        .comment { font-size: 80%; font-style: italic;}
        .entry > a { text-decoration: none; color: black; }
        .entry > a:hover { color: red; }
        .title { background: linear-gradient(90deg, #E0E0E0, #F0F0F0); border-radius: 6px; }
        .link, .title { width: 40%; }
        .title { text-transform: uppercase; }
        .logo { width: 5%; min-width: 5em; flex-grow: 0; background: #F0F0F0; }
        .logo img { width: 4em; height: 4em; }
        .stats { width: 10%; min-width: 10em; }
        .feed { font-size: 90%; }
        .actions { display: flex; flex-direction: row; align-content: flex-end; flex: wrap; width: 100vw; background: #F0F0F0; }
        .actions div:nth-of-type(1) { flex-grow: 1; }
        .actions div { margin: 0 0 0 2em; background: inherit; }
        .actions a { display: block; border: 1px solid #999; padding: 4px; color: black; text-decoration: none; border-radius: 5px; font-weight: bold; color: #222; background: white; }
        .actions a:hover { color: white; background: #222; }
        #delete { color: red; background: white; }
        #delete:hover { color: white; background: red; }
        #update-settings { background: inherit; }

        .loader { display: inline-block; width: 4em; height: 4em; }
        .loader:after { content: " "; display: block; width: 2em; height: 2em; margin: 8px; border-radius: 50%; border: 6px solid black; border-color: black transparent black transparent; animation: loader 1.2s linear infinite; }
        @keyframes loader { 0% { transform: rotate(0deg); } 100% { transform: rotate(360deg); } }

        #toast { position: absolute; width: 33%; left: 33%; bottom: 2em; padding: 2em; border-radius: 6px; background: #F0F0F0; font-size: 1.2em; font-family: sans-serif; z-index: 200; text-align: center;}
        #toast.success { border: 1px solid black; }
        #toast.error { border: 6px solid red; }

        legend { margin-left: 1.5em; font-size: 100%; border: 1px solid black; border-radius: 6px; background: white; padding: 0.1em 0.5em; }
        .error legend { border: 2px solid red; }
 
        @media all and (max-width: 500px) {
            .show { display: block; }
            .entry { width: 95% !important; }
            .last { margin-bottom: 1em; }
            .actions { flex-direction: column; align-items: flex-end; }
            .actions div { margin: 0.2em 0; }
        }
    </style>
</head>
<body>
    <div id="new-feed">
        <input id="url" name="feed" type="text" placeholder="Spotify show/episode url">
        <button id='new'>Add</button>
        <button id='opensettings' onclick="document.getElementById('settings').classList.remove('hidden'); return false;">Settings</button>
    </div>
    <div id="feed-list">
    <?php foreach($feeds as $url => ["title"=>$title,"image"=>$image,"episodes"=>$episodes,"last"=>$last,"max"=>$max,"keep"=>$keep,"feed"=>$feed]): ?>
        <div class="show" id="<?php echo basename($url) ?>">
            <div class="overlay loader <?php echo (array_key_exists($url, $ACTIVE)) ? "" : "hidden" ?>"></div>
            <div class="entry title"><?=$title?></div>
            <div class="entry logo"><img src="<?=$feed."/".SHOW_IMAGE?>"/></div>
            <div class="entry link"><a target="_blank" href="<?=$url?>"><?=htmlspecialchars($url)?></a></div>
            <div class="entry stats" id="stats-<?php echo basename($url) ?>"><?=$last?>&nbsp;(<?=$episodes?>)</div>
            <div class="entry feed"><a target="_blank" href="<?=$feed?>"><svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor" class="bi bi-rss" viewBox="0 0 16 16"><path d="M14 1a1 1 0 0 1 1 1v12a1 1 0 0 1-1 1H2a1 1 0 0 1-1-1V2a1 1 0 0 1 1-1h12zM2 0a2 2 0 0 0-2 2v12a2 2 0 0 0 2 2h12a2 2 0 0 0 2-2V2a2 2 0 0 0-2-2H2z"/><path d="M5.5 12a1.5 1.5 0 1 1-3 0 1.5 1.5 0 0 1 3 0zm-3-8.5a1 1 0 0 1 1-1c5.523 0 10 4.477 10 10a1 1 0 1 1-2 0 8 8 0 0 0-8-8 1 1 0 0 1-1-1zm0 4a1 1 0 0 1 1-1 6 6 0 0 1 6 6 1 1 0 1 1-2 0 4 4 0 0 0-4-4 1 1 0 0 1-1-1z"/></svg>&nbsp;&nbsp;<?=$feed?></a></div>
            <div class="entry last actions"><div></div><div>sync&nbsp;<select id="max-<?=basename($url)?>" onchange="showUpdate(this,'<?=$url?>','max')"><?php foreach([0,1,2,3,4,5] as $i) { printf("<option value='%d'>%d</option>",$i, $i);}?></select></div><div>keep&nbsp;<select id="keep-<?=basename($url)?>" onchange="showUpdate(this,'<?=$url?>','keep')"><?php foreach([1,2,5,10,25,50,100,250,1000,2500] as $i) { printf("<option value='%d'>%d</option>",$i, $i);}?></select>&nbsp;</div><div><a id="delete" onclick="confirmDelete('<?=$title?>','<?=$url?>')">delete</a></div><div><a id="refresh" onclick="refresh('<?=$url?>')">refresh</a></div></div>
        </div>
    <?php endforeach; ?>
    </div>
    <div id="settings" class="hidden">
        <div id="settings-header">
            <button id='closesettings' onclick="document.getElementById('settings').classList.add('hidden'); return false;">Close</button>
        </div>
        <fieldset id="credentials" class="setgroup">
            <legend>Spotify login</legend>
            <div class="setitem"><input id="username" name="username" type="text" placeholder="Spotify username">&nbsp;
            <input id="password" name="password" type="password" placeholder="Spotify password"></div>
            <div class="setitem"><button id="do-login" name="do-login">Login to Spotify</button></div>
        </fieldset>
        <fieldset id="scheduler" class="setgroup">
            <legend>Feed update schedule</legend>
            <div class="setitem comment"><span>There are two ways to keep your shows up to date:<ol><li>using the native scheduler</li><li>through webcron: <b><?=$SPODCAST_URL.'?action=update_shows'?></b></li></ol>Enable updates through the native scheduler by clicking the button. If this does not work - e.g. because the web server user is not allowed to add cron jobs or because the web server runs in a Docker container which does not support such jobs - you can point any web client (curl or wget work fine here) at the webcron endpoint (see #2) to initiate an update run. Configure a user cron job for curl/wget to access the refresh endpoint at the desired times for a user experience similar to using the native scheduler.</span>  </div>
            <div class="setitem"><button id="update-enable" name="update-enable">Enable scheduled updates</button></div>
            <div id="update-settings">
            <div class="setitem">Update feed&nbsp;<select id="update-rate"><option value="1">1</option><option value="2">2</option><option value="3">3</option><option value="4">4</option><option value="6">6</option><option value="8">8</option><option value="12">12</option><option value="24">24</option></select>&nbsp;times per day starting around&nbsp;<select id="update-start"><?php for ($i=0; $i<=24; $i++) { printf("<option value='%d'>%02d:00</option>",$i, $i);}?></select></div>
            <div class="setitem comment">Updates start 5 to 25 minutes after the hour</div>
            <div class="setitem"><button id="do-update" name="do-update">Update schedule</button>
            <button id="update-disable" name="update-disable">Disable updates</button></div>
            </fieldset>
        <fieldset id="transcode" class="setgroup">
            <legend>Transcode</legend>
            <div class="setitem comment"><span>Some devices - mainly Apple iOS - do not support open audio codecs like those used by Spotify. For such devices Spodcast can transcode ogg streams to mp3. This is an expensive operation which can take a substantial amount of time, especially on less powerful hardware. Only enable this option when there are no other options.</span></div>
            <div class="setitem"><button id="transcode-enable" name="transcode-enable">Enable transcoder</button><button id="transcode-disable" name="transcode-disable">Disable transcoder</button></div>
        </fieldset>
        <fieldset id="logging" class="setgroup">
            <legend>Logging</legend>
            <div class="setitem comment"><span>The default log level shows which episodes have been downloaded. Change the log level to get more (or less) elaborate information on what Spodcast is up to. Log messages are reported in plain text format through scheduled updates and json-encoded through the webcron endpoint.</span></div>
            <div class="setitem"><select id="log-level" onchange="logUpdate(this)"><?php foreach([['CRITICAL','Only show critical errors'],['ERROR','Show error messages'],['WARNING','Show which episodes have been downloaded'],['INFO','Show downloaded as well as skipped episodes'],['DEBUG','Show detailed internal information for debugging purposes']] as $s) { printf("<option value='%s'>%s</option>",$s[0], $s[1]);}?></select></div>
        </fieldset>


        </div>
    </div>
    <div id="error" class="hidden">
    <fieldset id="errormessage" class="error">
        <legend>Error: <?=$ERROR_MESSAGE?></legend>
        <div class="errorline"><pre><?=$ERROR_DETAILS?></pre></div>
        <div class="errorline"><button id="hide-error" onclick="window.location.href='.'; return false;">Close this message</button></div>
    </fieldset>
    </div>
    <div class="hidden" id="toast"></div>
    <script>
    function xhr(type, url, data, options) {
        options = options || {};
        var request = new XMLHttpRequest();
        request.open(type, url, true);
        if(type === "POST") {
            request.setRequestHeader('Content-type', 'application/x-www-form-urlencoded');
        }
        request.onreadystatechange = function () {
            if (this.readyState === 4) {
                if (this.status >= 200 && this.status < 400) {
                    options.success && options.success(parse(this.responseText));
                } else {
                    options.error && options.error(this.status);
                }
            }
        };
        request.send(data);
    }
    function parse(text) {
        try {
            return JSON.parse(text);
        } catch(e) {
            return text;
        }
    }
    function basename(path) {
        return path.replace(/.*\//, '');
    }
    function newfeed() {
        xhr("POST", '?action=new', 'url=' + encodeURI(document.getElementById('url').value), {success:addShow,error:removeShade});
    }
    function addShow(info) {
        title="Adding show (" + info.url + "), please wait...";
        show=document.createElement('div'); show.setAttribute('class', 'show placeholder'); show.setAttribute('id', info.id);
        shade=document.createElement('div'); shade.setAttribute('class', 'overlay');
        logo=document.createElement('div'); logo.setAttribute('class', 'entry logo loader');
        title=document.createElement('div'); title.setAttribute('class', 'entry title');
        feedlist=document.getElementById('feed-list');
        show.appendChild(shade); show.appendChild(title); show.appendChild(logo);
        title.innterHTML = title;
        first=feedlist.firstElementChild;
        feedlist.insertBefore(show, first);
        waitUpdate(info);
    }
    function login() {
        xhr("POST", '?action=login', 'username=' + document.getElementById('username').value + '&password=' + document.getElementById('password').value, {success:showToast, error:showToast});
    }
    function refresh(url) {
        unhide(basename(url));
        id = basename(url);
        max=document.getElementById('max-' + id).value;
        keep=document.getElementById('keep-' + id).value;
        xhr("POST", '?action=refresh', 'url=' + encodeURI(url) + '&max=' + max + '&keep=' + keep, {success:wait, error:removeShade});
    }
    function waitUpdate(info) {
        if (info.status == 'READY' || info.status == 'ERROR') {
            window.location.href = '';
        }
        if (info.show != null) {
            document.getElementById(info.id).childNodes.item(1).innerHTML = "Adding show: " + info.show.title + " (" + info.show.episodes + " episodes of 3 downloaded)...";
        } else {
            document.getElementById(info.id).childNodes.item(1).innerHTML = "Adding show: " + info.url + "...";
        }
        setTimeout(function() {
                xhr("POST", '?action=status', 'url=' + encodeURI(info.url), {success:waitUpdate, error:removeShade});
                }, 2000);
    }
    function wait(info) {
        if (info.status == 'READY' || info.status == 'ERROR') {
            document.getElementById('stats-' + info.id).innerHTML = info.show.last + '&nbsp;(' + info.show.episodes + ')';
            hide(info.id);
        } else { 
            if (info.show != null) {
                document.getElementById('stats-' + info.id).innerHTML = info.show.last + '&nbsp;(' + info.show.episodes + ')';
            }
            setTimeout(function() {
                    xhr("POST", '?action=status', 'url=' + encodeURI(info.url), {success:wait, error:removeShade});
                    }, 2000);
        }
    }
    function removeShade(info) {
        hide(info.id);
    }
    function hide(id) {
        document.getElementById(id).firstElementChild.classList.add('hidden');
    }
    function unhide(id) {
        document.getElementById(id).firstElementChild.classList.remove('hidden');
    }
    function schedule(enable) {
        xhr("POST", '?action=schedule', 'enable=' + enable + '&rate=' + document.getElementById('update-rate').value + '&start=' + document.getElementById('update-start').value, {success:showToast, error:showToast});
    }
    function transcode(enable) {
        xhr("POST", '?action=transcode', 'enable=' + enable, {success:showToast, error:showToast});
    }
    function logUpdate(e) {
        xhr("POST", '?action=logging', 'level=' + e.value, {success:showToast, error:showToast});
    }
    function selectElement(id, val) {
        document.getElementById(id).value=val;
    }
    function showUpdate(e,url,field) {
        id = basename(url);
        max=document.getElementById('max-' + id);
        keep=document.getElementById('keep-' + id);
        maxval = parseInt(max.value);
        keepval = parseInt(keep.value);
        if(maxval > keepval) {
            keeper = (maxval > 2) ? 5 : 2;
            if (confirm("Sync count " + maxval + " is higher than Keep count " + keepval + "\n\n" +
                "It does not make sense to refresh more expisodes than are kept in cache." +
                "Press OK to continue, Keep will be set to " + keeper + ". Otherwise press Cancel")) {
                keep.value = keeper;
            } else {
                keep.value = keepval;
                max.value = maxval;
                return false;
            }
        }
        xhr("POST", '?action=update', 'url=' + encodeURI(url) + '&field=' + field + '&value=' + e.value, {success:showToast, error:showToast});
    }
    function confirmDelete(title, url) {
        if (confirm('You are about to delete the following feed:\n\nTitle: ' + title + '\nURL: ' + url + '\n\nAre you sure you want to delete this feed?')) {
            xhr("POST", '?action=delete', 'url=' + encodeURI(url), {success:deleteShow, error:showToast});
        }
        return false;
    }
    function deleteShow(info) {
        document.getElementById(info.id).remove();
        showToast(info);
    }
    function showToast(info) {
        var toast=document.getElementById('toast');
        if (info.status == 'ERROR') {
            toast.classList.remove('success');
            toast.classList.add('error');
        } else {
            toast.classList.remove('error');
            toast.classList.add('success');
        }
        toast.innerHTML=info.result;
        toast.classList.remove("hidden");
        window.setTimeout(function() {
            toast = document.getElementById('toast');
            toast.classList.add("hidden");
            toast.classList.replace('error', 'success');
            toast.innerHTML="";
        }, 2000);
    }

    document.getElementById('new').onclick = function() {
        newfeed();
    };
    document.getElementById('do-login').onclick = function() {
        login();
    };
    document.getElementById('do-update').onclick = function() {
        schedule(true);
    };
    document.getElementById('url').onkeydown = function(e) {
        if(e.keyCode == 13) {
            newfeed();
        }
    };

    var schedule_enable = document.getElementById('update-enable');
    var schedule_disable = document.getElementById('update-disable');
    var schedule_entries = document.getElementById('update-settings');
    var transcode_enable = document.getElementById('transcode-enable');
    var transcode_disable = document.getElementById('transcode-disable');
    schedule_entries.style.display = '<?php echo ($UPDATE_ENABLED ? 'initial' : 'none');?>';
    schedule_enable.style.display = '<?php echo ($UPDATE_ENABLED ? 'none' : 'initial');?>';
    transcode_enable.style.display = '<?php echo ($TRANSCODE_ENABLED ? 'none' : 'initial');?>';
    transcode_disable.style.display = '<?php echo ($TRANSCODE_ENABLED ? 'initial' : 'none');?>';

    schedule_enable.onclick = function() {
        schedule_entries.style.display = 'initial';
        schedule_enable.style.display = 'none';
        schedule(true);
    };
    schedule_disable.onclick = function() {
        schedule_entries.style.display = 'none';
        schedule_enable.style.display = 'initial';
        schedule(false);
    };
    transcode_enable.onclick = function() {
        transcode_enable.style.display = 'none';
        transcode_disable.style.display = 'initial';
        transcode(true);
    };
    transcode_disable.onclick = function() {
        transcode_enable.style.display = 'initial';
        transcode_disable.style.display = 'none';
        transcode(false);
    };
    selectElement('update-rate', '<?=$UPDATE_RATE?>');
    selectElement('update-start', '<?=$UPDATE_START?>');
    selectElement('log-level', '<?=$LOG_LEVEL?>');
    <?php foreach($feeds as $url => ["keep"=>$keep,"max"=>$max]) {
        printf("    selectElement('keep-%s', '%d');\n", basename($url), $keep);
        printf("    selectElement('max-%s', '%d');\n", basename($url), $max);
    }?>
    <?php if ($ERROR_MESSAGE != null) {
        echo "    document.getElementById('error').classList.remove('hidden');";
    }?>
    </script>
</body>
</html>'''




import json
import os
import pkg_resources
import sys
from typing import Any

CONFIG_FILE_PATH = '../spodcast.json'

CONFIG_DIR = 'CONFIG_DIR'
CONFIG_PATH = 'CONFIG_PATH'
VERSION = 'VERSION'
VERSION_STR = pkg_resources.require("Spodcast")[0].version

ROOT_PATH = 'ROOT_PATH'
SKIP_EXISTING_FILES = 'SKIP_EXISTING_FILES'
CHUNK_SIZE = 'CHUNK_SIZE'
DOWNLOAD_REAL_TIME = 'DOWNLOAD_REAL_TIME'
LANGUAGE = 'LANGUAGE'
CREDENTIALS_LOCATION = 'CREDENTIALS_LOCATION'
RETRY = 'RETRY'
MAX_EPISODES = 'MAX_EPISODES'
LOG_LEVEL = 'LOG_LEVEL'
RSS_FEED = 'RSS_FEED'
TRANSCODE = 'TRANSCODE'

CONFIG_VALUES = {
    ROOT_PATH:            { 'default': '../Spodcast/',
                            'type': str,
                            'arg': '--root-path',
                            'help': 'set root path for podcast cache' },
    SKIP_EXISTING_FILES:  { 'default': 'True',
                            'type': bool,
                            'arg': '--skip-existing',
                            'help': '[yes|no] skip files with the same name and size. Defaults to "yes".' },
    RETRY:                { 'default': 5,
                            'type': int,
                            'arg': '--retry',
                            'help': 'retry count for Spotify API access' },
    MAX_EPISODES:         { 'default': 1000,
                            'type': int,
                            'arg': '--max-episodes',
                            'help': 'number of episodes to download' },
    CHUNK_SIZE:           { 'default': 50000,
                            'type': int,
                            'arg': '--chunk-size',
                            'help': 'download chunk size' },
    DOWNLOAD_REAL_TIME:   { 'default': 'False',
                            'type': bool,
                            'arg': '--download-real-time',
                            'help': 'simulate streaming client' },
    LANGUAGE:             { 'default': 'en',
                            'type': str,
                            'arg': '--language',
                            'help': 'preferred content language' },
    CREDENTIALS_LOCATION: { 'default': 'credentials.json',
                            'type': str,
                            'arg': '--credentials-location',
                            'help': 'path to credentials file. If a relative path is used the file will be stored in the same directory as the configuration file (-c /path/to/config.json -> /path/to).' },
    RSS_FEED:             { 'default': 'True',
                            'type': bool,
                            'arg': '--rss-feed',
                            'help': '[yes|no] add a (php) RSS feed server and related metadata for feed. To manage feeds, point a web server at the spodcast root path as configured using --root-path. Defaults to "yes".' },
    TRANSCODE:            { 'default': 'False',
                            'type': bool,
                            'arg': '--transcode',
                            'help': '[yes|no] transcode ogg/vorbis to mp3 (where applicable) - only needed for devices which do not support open formats (e.g. iOS). Defaults to "no".' },
    LOG_LEVEL:            { 'default': 'warning',
                            'type': str,
                            'arg': '--log-level',
                            'help': 'log level (debug/info/warning/error/critical)' }
}

class Config:
    Values = {}

    @classmethod
    def load(cls, args) -> None:
        app_dir = os.path.dirname(__file__)
        dump_config=False

        config_fp = CONFIG_FILE_PATH
        if args.config_location:
            config_fp = args.config_location

        true_config_file_path = os.path.join(app_dir, config_fp)

        # Load config from spodcast.json

        if not os.path.exists(true_config_file_path):
            dump_config=True
            os.makedirs(os.path.dirname(true_config_file_path), exist_ok=True)
            cls.Values = cls.get_default_json()
        else:
            with open(true_config_file_path, encoding='utf-8') as config_file:
                jsonvalues = json.load(config_file)
                cls.Values = {}
                for key in CONFIG_VALUES:
                    if key in jsonvalues:
                        cls.Values[key] = cls.parse_arg_value(key, jsonvalues[key])

        # Add default values for missing keys

        for key in CONFIG_VALUES:
            if key not in cls.Values:
                cls.Values[key] = cls.parse_arg_value(key, CONFIG_VALUES[key]['default'])
                dump_config=True

        # Override config from commandline arguments

        for key in CONFIG_VALUES:
            if key.lower() in vars(args) and vars(args)[key.lower()] is not None:
                cls.Values[key] = cls.parse_arg_value(key, vars(args)[key.lower()])

        # dump current config to config file

        if dump_config:
            with open(true_config_file_path, 'w', encoding='utf-8') as config_file:
                json.dump(cls.get_config_json(), config_file, indent=4)

        # these values should not be overriden

        cls.Values[CONFIG_DIR] = os.path.dirname(true_config_file_path)
        cls.Values[CONFIG_PATH] = str(true_config_file_path)
        cls.Values[VERSION] = str(VERSION_STR)

    @classmethod
    def get_default_json(cls) -> Any:
        r = {}
        for key in CONFIG_VALUES:
            r[key] = CONFIG_VALUES[key]['default']
        return r

    @classmethod
    def get_config_json(cls) -> Any:
        r = {}
        for key in CONFIG_VALUES:
            r[key]=cls.Values[key]
        return r

    @classmethod
    def parse_arg_value(cls, key: str, value: Any) -> Any:
        if type(value) == CONFIG_VALUES[key]['type']:
            return value
        if CONFIG_VALUES[key]['type'] == str:
            return str(value)
        if CONFIG_VALUES[key]['type'] == int:
            return int(value)
        if CONFIG_VALUES[key]['type'] == bool:
            if str(value).lower() in ['yes', 'true', '1']:
                return True
            if str(value).lower() in ['no', 'false', '0']:
                return False
            raise ValueError("Not a boolean: " + value)
        raise ValueError("Unknown Type: " + value)

    @classmethod
    def get(cls, key: str) -> Any:
        return cls.Values.get(key)

    @classmethod
    def get_config_dir(cls) -> str:
        return cls.get(CONFIG_DIR)

    @classmethod
    def get_root_path(cls) -> str:
        return os.path.join(os.path.dirname(__file__), cls.get(ROOT_PATH))

    @classmethod
    def get_skip_existing_files(cls) -> bool:
        return cls.get(SKIP_EXISTING_FILES)

    @classmethod
    def get_chunk_size(cls) -> int:
        return cls.get(CHUNK_SIZE)

    @classmethod
    def get_language(cls) -> str:
        return cls.get(LANGUAGE)

    @classmethod
    def get_download_real_time(cls) -> bool:
        return cls.get(DOWNLOAD_REAL_TIME)

    @classmethod
    def get_credentials_location(cls) -> str:
        credentials_location = cls.get(CREDENTIALS_LOCATION)
        return credentials_location if os.path.isabs(credentials_location) else os.path.join(cls.get(CONFIG_DIR), cls.get(CREDENTIALS_LOCATION))

    @classmethod
    def get_retry(cls) -> int:
        return cls.get(RETRY)

    @classmethod
    def get_max_episodes(cls) -> int:
        return cls.get(MAX_EPISODES)

    @classmethod
    def get_rss_feed(cls) -> bool:
        return cls.get(RSS_FEED)

    @classmethod
    def get_transcode(cls) -> bool:
        return cls.get(TRANSCODE)

    @classmethod
    def get_log_level(cls) -> str:
        return str(cls.get(LOG_LEVEL)).upper()

    @classmethod
    def get_bin_path(cls) -> str:
        return str(sys.argv[0])

    @classmethod
    def get_config_path(cls) -> str:
        return cls.get(CONFIG_PATH)

    @classmethod
    def get_version_str(cls) -> str:
        return cls.get(VERSION)

    @classmethod
    def get_version_int(cls) -> int:
        return int(cls.get(VERSION).replace('.',''))



from getpass import getpass
import glob
import hashlib
from logging import Logger
import logging
import os
import os.path
import random
import requests
import sys
import time

from librespot.audio.decoders import VorbisOnlyAudioQuality
from librespot.core import Session

from spodcast.config import Config
from spodcast.const import CREDENTIALS_PREFIX, TYPE, USER_READ_EMAIL, OFFSET, LIMIT
from spodcast.feedgenerator import RSS_FEED_FILE_NAME, RSS_INDEX_CODE, get_index_version

class Spodcast:    
    SESSION: Session = None
    DOWNLOAD_QUALITY = None
    CONFIG: Config = Config()
    LOG: Logger = None

    def __init__(self, args):
        Spodcast.CONFIG.load(args)
        logging.basicConfig(level=Spodcast.CONFIG.get_log_level())
        log = logging.getLogger(__name__)
        Spodcast.LOG = log
        log.debug("args: %s", args)
        if args.prepare_feed is True:
            root_path=Spodcast.CONFIG.get_root_path()
            os.makedirs(root_path, exist_ok=True)
            if os.path.exists(root_path):
                index_file_name = os.path.join(root_path, RSS_FEED_FILE_NAME)
                if not os.path.isfile(index_file_name) or int(get_index_version(index_file_name)) < Spodcast.CONFIG.get_version_int():
                    rss_file = open(index_file_name, "w")
                    rss_file.write(RSS_INDEX_CODE(Spodcast.CONFIG.get_bin_path(), os.path.basename(Spodcast.CONFIG.get_config_path()), Spodcast.CONFIG.get_version_str()))
                    rss_file.close()
            else:
                sys.exit(f"Can not create root path {root_path}")

        if args.login:
            filename = args.login
            if os.path.exists(filename):
                with open(filename, 'r', encoding='utf-8') as file:
                    for line in file.readlines():
                        Spodcast.account(line.strip())
                sys.exit(0)
            else:
                sys.exit(f"Can not read username/password file {filename}")

        Spodcast.login()

    @classmethod
    def account(cls, line):
        cred_directory = Config.get_config_dir()
        if os.path.isdir(cred_directory):
            (username,password) = line.split()
            cred_filename = CREDENTIALS_PREFIX + "-" + hashlib.md5(username.encode('utf-8'),usedforsecurity=False).hexdigest() + ".json"
            cred_location = os.path.join(cred_directory, cred_filename)
            conf = Session.Configuration.Builder().set_stored_credential_file(cred_location).build()
            session = Session.Builder(conf).user_pass(username, password).create()
            if not session.is_valid():
                Spodcast.LOG.error("Invalid username/password for username " + username);
             
    @classmethod
    def login(cls):
        cred_directory = Config.get_config_dir()
        credfiles = glob.glob(os.path.join(cred_directory, CREDENTIALS_PREFIX) + "-*.json")
        if credfiles:
            random.shuffle(credfiles)
            for credfile in credfiles:
                try:
                    cred_location = os.path.join(cred_directory, credfile)
                    conf = Session.Configuration.Builder().set_stored_credential_file(cred_location).set_store_credentials(False).build()
                    session = Session.Builder(conf).stored_file().create()
                    if session.is_valid():
                        cls.SESSION = session
                        return
                    else:
                        Spodcast.LOG.warning(f"Invalid credentials in {cred_location}")
                except RuntimeError:
                    Spodcast.LOG.error("RuntimeError")
                    pass
                
        cred_location = Config.get_credentials_location()

        if os.path.isfile(cred_location):
            try:
                conf = Session.Configuration.Builder().set_stored_credential_file(cred_location).set_store_credentials(False).build()
                cls.SESSION = Session.Builder(conf).stored_file().create()
                return
            except RuntimeError:
                pass
        while True:
            user_name = ''
            while len(user_name) == 0:
                user_name = input('Username: ')
            password = getpass()
            try:
                conf = Session.Configuration.Builder().set_stored_credential_file(cred_location).build()
                cls.SESSION = Session.Builder(conf).user_pass(user_name, password).create()
                return
            except RuntimeError:
                pass

    @classmethod
    def get_content_stream(cls, content_id, quality):
        return cls.SESSION.content_feeder().load(content_id, VorbisOnlyAudioQuality(quality), False, None)

    @classmethod
    def __get_auth_token(cls):
        return cls.SESSION.tokens().get_token(USER_READ_EMAIL).access_token

    @classmethod
    def get_auth_header(cls):
        return {
            'Authorization': f'Bearer {cls.__get_auth_token()}',
            'Accept-Language': f'{cls.CONFIG.get_language()}'
        }

    @classmethod
    def get_auth_header_and_params(cls, limit, offset):
        return {
            'Authorization': f'Bearer {cls.__get_auth_token()}',
            'Accept-Language': f'{cls.CONFIG.get_language()}'
        }, {LIMIT: limit, OFFSET: offset}

    @classmethod
    def invoke_url_with_params(cls, url, limit, offset, **kwargs):
        headers, params = cls.get_auth_header_and_params(limit=limit, offset=offset)
        params.update(kwargs)
        return requests.get(url, headers=headers, params=params).json()

    @classmethod
    def invoke_url(cls, url, tryCount=0):
        headers = cls.get_auth_header()
        Spodcast.LOG.debug(headers)
        response = requests.get(url, headers=headers)
        responsetext = response.text
        responsejson = response.json()

        if 'error' in responsejson:
            if tryCount < (cls.CONFIG.get_retry() - 1):
                Spodcast.LOG.warning(f"Spotify API Error (try {tryCount + 1}) ({responsejson['error']['status']}): {responsejson['error']['message']}")
                time.sleep(5)
                return cls.invoke_url(url, tryCount + 1)

            Spodcast.LOG.error(f"Spotify API Error ({responsejson['error']['status']}): {responsejson['error']['message']}")

        return responsetext, responsejson



import argparse
import pkg_resources

from spodcast.app import client
from spodcast.config import CONFIG_VALUES

def main():
    parser = argparse.ArgumentParser(prog='spodcast', description='A caching Spotify podcast to RSS proxy.')
    parser.add_argument('-c', '--config-location',
                        type=str,
                        help='Specify the spodcast.json location')

    parser.add_argument('-p', '--prepare-feed',
                       action='store_true',
                       help='Installs RSS feed server code in ROOT_PATH.')

    parser.add_argument('-v', '--version',
                       action='version',
                       version = '%(prog)s ' + pkg_resources.require("Spodcast")[0].version)

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('urls',
                       type=str,
                       # action='extend',
                       default='',
                       nargs='*',
                       help='Download podcast episode(s) from a url. Can take multiple urls.')

    group.add_argument('-l', '--login',
                       type=str,
                       help='Reads username and password from file passed as argument and stores credentials for later use.')

    for configkey in CONFIG_VALUES:
        parser.add_argument(CONFIG_VALUES[configkey]['arg'],
                            type=str,
                            default=None,
                            help=CONFIG_VALUES[configkey]['help'])

    parser.set_defaults(func=client)

    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
    main()



TYPE = 'type'
LIMIT = 'limit'
OFFSET = 'offset'
CREDENTIALS_PREFIX = 'spodcast-cred'
USER_READ_EMAIL = 'user-read-email'
FILE_EXISTS = -1
OPEN_SPOTIFY_URL = 'open.spotify.com'
IMAGE_CDN = lambda image_id_hex: f'https://i.scdn.co/image/{image_id_hex}'



import json
import logging
import os
import time
from datetime import datetime
from html import escape
import urllib.parse

import base62
from base62 import CHARSET_INVERTED
import ffmpeg

from librespot import util
from librespot.metadata import ShowId, EpisodeId
from librespot.core import ApiClient

from spodcast.const import FILE_EXISTS, IMAGE_CDN
from spodcast.feedgenerator import RSS_FEED_CODE, RSS_FEED_FILE_NAME, RSS_FEED_SHOW_INDEX, RSS_FEED_INFO_EXTENSION, RSS_FEED_SHOW_IMAGE, RSS_FEED_VERSION, get_index_version
from spodcast.utils import clean_filename, uri_to_url
from spodcast.spodcast import Spodcast

log = logging.getLogger(__name__)


def hex_to_spotify_id(hex_id):
    return base62.encodebytes(util.hex_to_bytes(hex_id), CHARSET_INVERTED)


def get_show_info(show_id_hex):
    log.info("Fetching show information...")
    show_id = ShowId.from_hex(show_id_hex)
    uri = f'spotify:show:{hex_to_spotify_id(show_id_hex)}'
    info = Spodcast.SESSION.api().get_metadata_4_show(show_id)
    link = uri_to_url(uri)
    description = info.description
    image = IMAGE_CDN(util.bytes_to_hex(info.cover_image.image[1].file_id))

    return link, description, image


def get_episode_info(episode_id_hex):
    log.info("Fetching episode information...")
    episode_id = EpisodeId.from_hex(episode_id_hex)
    uri = f'spotify:episode:{hex_to_spotify_id(episode_id_hex)}'
    info = Spodcast.SESSION.api().get_metadata_4_episode(episode_id)
    podcast_name = info.show.name
    podcast_id = util.bytes_to_hex(info.show.gid)
    episode_name = info.name
    duration_ms = info.duration
    description = info.description
    external_url = info.external_url if info.external_url else None
    pt = info.publish_time
    release_date = f'{pt.year}-{pt.month}-{pt.day}T{pt.hour}:{pt.minute}:00Z'

    return podcast_name, podcast_id, duration_ms, episode_name, description, release_date, uri, external_url


def get_episodes(show_id):
    info = Spodcast.SESSION.api().get_metadata_4_show(show_id)
    episodes = info.episode
    episodes.sort(key = lambda x: datetime.strptime(f'{x.publish_time.year}-{x.publish_time.month}-{x.publish_time.day}T{x.publish_time.hour}:{x.publish_time.minute}:00Z', "%Y-%m-%dT%H:%M:%SZ"), reverse=True)

    return [util.bytes_to_hex(episode.gid) for episode in episodes]


def download_file(url, filepath):
    import functools
    import pathlib
    import shutil
    import requests

    mimetype = "audio/mpeg"

    r = requests.get(url, stream=True, allow_redirects=True)
    if r.status_code != 200:
        r.raise_for_status()  # Will only raise for 4xx codes, so...
        log.error(f"Request to {url} returned status code {r.status_code}")
        return

    file_size = int(r.headers.get('Content-Length', 0))

    if (
        os.path.isfile(filepath)
        and abs(file_size - os.path.getsize(filepath)) < 1000
        and Spodcast.CONFIG.get_skip_existing_files()
    ):
        return filepath, FILE_EXISTS, mimetype

    log.info("Downloading file")
    r.raw.read = functools.partial(r.raw.read, decode_content=True)
    with open(filepath, "wb") as file:
        shutil.copyfileobj(r.raw, file)


    return filepath, os.path.getsize(filepath), mimetype


def download_stream(stream, filepath):
    size = stream.input_stream.size

    mp3_filepath = os.path.splitext(filepath)[0] + ".mp3"
    mimetype = "audio/ogg"

    if (
        # "FILE SIZE CHECK TEMPORARILY OUT OF ORDER"
        # Need to find a way to get decrypted content size
        # from Spotify to enable file size checks, for now
        # this only checks for the presence of a file with
        # the same name. To recover from failed downloads
        # simply remove incomplete files
        #
        #((os.path.isfile(filepath)
        #and abs(size - os.path.getsize(filepath)) < 1000)
        (os.path.isfile(filepath)
        or (Spodcast.CONFIG.get_transcode()
        and os.path.isfile(mp3_filepath)))
        and Spodcast.CONFIG.get_skip_existing_files()
    ):
        return filepath, FILE_EXISTS, mimetype
 
    log.info("Downloading stream")
    time_start = time.time()
    downloaded = 0
    with open(filepath, 'wb') as file:
        for _ in range(int(size / Spodcast.CONFIG.get_chunk_size()) + 1):
            data = stream.input_stream.stream().read(Spodcast.CONFIG.get_chunk_size())
            file.write(data)
            downloaded += len(data)
            if Spodcast.CONFIG.get_download_real_time():
                delta_real = time.time() - time_start
                delta_want = (downloaded / size) * (duration_ms/1000)
                log.debug(f"realtime enabled, waiting for {delta_real} seconds...")
                if delta_want > delta_real:
                    time.sleep(delta_want - delta_real)

    if Spodcast.CONFIG.get_transcode():
        log.info("transcoding ogg->mp3")
        transcoder = ffmpeg.input(filepath)
        transcoder = ffmpeg.output(transcoder, mp3_filepath)
        ffmpeg.run(transcoder, quiet=True)
        file.close()
        os.unlink(filepath)
        filepath = mp3_filepath
        downloaded = os.path.getsize(filepath)
        mimetype = "audio/mpeg"

    return filepath, downloaded, mimetype


def download_episode(episode_id) -> None:
    try:
        podcast_name, podcast_id, duration_ms, episode_name, description, release_date, uri, download_url = get_episode_info(episode_id)

        if podcast_name is None:
            log.warning('Skipping episode (podcast NOT FOUND)')
        elif episode_name is None:
            log.warning('Skipping episode (episode NOT FOUND)')
        else:
            filename = clean_filename(podcast_name + ' - ' + episode_name)
            show_directory = os.path.realpath(os.path.join(Spodcast.CONFIG.get_root_path(), clean_filename(podcast_name) + '/'))
            os.makedirs(show_directory, exist_ok=True)

            if download_url is None:
                episode_stream_id = EpisodeId.from_hex(episode_id)
                stream = Spodcast.get_content_stream(episode_stream_id, Spodcast.DOWNLOAD_QUALITY)
                basename = f"{filename}.ogg"
                filepath = os.path.join(show_directory, basename)
                path, size, mimetype = download_stream(stream, filepath)
                basename = os.path.basename(path) # may have changed due to transcoding
            else:
                basename=f"{filename}.mp3"
                filepath = os.path.join(show_directory, basename)
                path, size, mimetype = download_file(download_url, filepath)

            if size == FILE_EXISTS:
                log.info(f"Skipped {podcast_name}: {episode_name}")
            else:
                log.warning(f"Downloaded {podcast_name}: {episode_name}")

                if Spodcast.CONFIG.get_rss_feed():
                    episode_info = {
                            "mimetype": mimetype,
                            "medium": "audio",
                            "duration": int(duration_ms/1000),
                            "date": time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.strptime(release_date, "%Y-%m-%dT%H:%M:%SZ")),
                            "title": escape(episode_name), "guid": uri, "description": escape(description),
                            "filename": urllib.parse.quote(basename),
                            "size": int(size) }
                    info_file = open(os.path.join(show_directory, f"{basename}.{RSS_FEED_INFO_EXTENSION}"), "w")
                    info_file.write(json.dumps(episode_info))
                    info_file.close()

            if Spodcast.CONFIG.get_rss_feed():
                show_index_file_name = os.path.join(show_directory, f"{RSS_FEED_SHOW_INDEX}.{RSS_FEED_INFO_EXTENSION}")
                if not os.path.isfile(show_index_file_name) or int(get_index_version(show_index_file_name)) < Spodcast.CONFIG.get_version_int():
                    podcast_link, podcast_description, podcast_image = get_show_info(podcast_id)
                    show_info = {}
                    if os.path.isfile(show_index_file_name):
                        with open(show_index_file_name, encoding='utf-8') as file:
                            show_info = json.load(file)
                            file.close()
                    show_info["version"] = str(RSS_FEED_VERSION + Spodcast.CONFIG.get_version_str())
                    show_info["title"] = escape(podcast_name)
                    show_info["link"] = podcast_link
                    show_info["description"] = escape(podcast_description)
                    show_info["image"] = RSS_FEED_SHOW_IMAGE 
                    show_index_file = open(show_index_file_name, "w")
                    show_index_file.write(json.dumps(show_info))
                    show_index_file.close()

                show_image_name = os.path.join(show_directory, f"{RSS_FEED_SHOW_IMAGE}")
                if not os.path.isfile(show_image_name):
                    download_file(podcast_image, show_image_name)

                rss_file_name = os.path.join(show_directory, RSS_FEED_FILE_NAME)
                if not os.path.isfile(rss_file_name) or int(get_index_version(rss_file_name)) < Spodcast.CONFIG.get_version_int():
                    rss_file = open(rss_file_name, "w")
                    rss_file.write(RSS_FEED_CODE(Spodcast.CONFIG.get_version_str()))
                    rss_file.close()

    except ApiClient.StatusCodeException as status:
        log.warning("episode %s, StatusCodeException: %s", episode_id, status)



import re
import string
import unicodedata
from enum import Enum
from typing import List, Tuple

from spodcast.spodcast import Spodcast
from spodcast.const import OPEN_SPOTIFY_URL

valid_filename_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)

def regex_input_for_urls(search_input) -> Tuple[str, str, str, str, str, str]:
    episode_uri_search = re.search(
        r'^spotify:episode:(?P<EpisodeID>[0-9a-zA-Z]{22})$', search_input)
    episode_url_search = re.search(
        r'^(https?://)?open\.spotify\.com/episode/(?P<EpisodeID>[0-9a-zA-Z]{22})(\?si=.+?)?$',
        search_input,
    )

    show_uri_search = re.search(
        r'^spotify:show:(?P<ShowID>[0-9a-zA-Z]{22})$', search_input)
    show_url_search = re.search(
        r'^(https?://)?open\.spotify\.com/show/(?P<ShowID>[0-9a-zA-Z]{22})(\?si=.+?)?$',
        search_input,
    )
    if episode_uri_search is not None or episode_url_search is not None:
        episode_id_str = (episode_uri_search
                          if episode_uri_search is not None else
                          episode_url_search).group('EpisodeID')
    else:
        episode_id_str = None

    if show_uri_search is not None or show_url_search is not None:
        show_id_str = (show_uri_search
                       if show_uri_search is not None else
                       show_url_search).group('ShowID')
    else:
        show_id_str = None

    return episode_id_str, show_id_str


def clean_filename(filename, whitelist=valid_filename_chars, replace=' '):
    for r in replace:
        filename = filename.replace(r,'_')

    cleaned_filename = unicodedata.normalize('NFKD', filename).encode('ASCII', 'ignore').decode()
    cleaned_filename = ''.join(c for c in cleaned_filename if c in whitelist)
    return cleaned_filename

def uri_to_url(spotify_id):
    (spotify,sp_type,sp_id) = spotify_id.split(':')
    return f'https://{OPEN_SPOTIFY_URL}/{sp_type}/{sp_id}'



import logging

from itertools import islice
from librespot.audio.decoders import AudioQuality
from librespot.metadata import ShowId, EpisodeId

from spodcast.podcast import download_episode, get_episodes
from spodcast.utils import regex_input_for_urls
from spodcast.spodcast import Spodcast

log = logging.getLogger(__name__)

def client(args) -> None:
    Spodcast(args)
    Spodcast.DOWNLOAD_QUALITY = AudioQuality.NORMAL

    if args.urls:
        for spotify_url in args.urls:
            episode_id_str, show_id_str = regex_input_for_urls(spotify_url)
            log.debug(f"episode_id_str {episode_id_str}. show_id_str {show_id_str}")
            if episode_id_str is not None:
                episode_id = EpisodeId.from_base62(episode_id_str).hex_id()
                log.debug("episode_id: %s", episode_id)
                download_episode(episode_id)
            elif show_id_str is not None:
                show_id = ShowId.from_base62(show_id_str)
                log.debug("show_id: %s", show_id)
                for episode_id in islice(get_episodes(show_id), Spodcast.CONFIG.get_max_episodes()):
                    log.debug("episode_id: %s", episode_id)
                    download_episode(episode_id)
